#!/bin/sh
#
# chrec - Record configuration changes for hosts.
#
# Copyright (c) 2006-2007 Occam's Razor. All rights reserved.
#
# See the LICENSE file distributed with this code for restrictions on its use
# and further distribution.
# Original distribution available at <http://www.occam.com/tools/>.
#
# $Id: chrec.sh,v 2.2 2007/06/06 01:15:33 leonvs Exp $
#
# TODO
#

PATH=/usr/xpg4/bin:/usr/bin:/bin

usage () {
	echo "$VERSION
usage:	$progname { -h | -V }
	$progname [-u user] [-D date] [-f filename] hostname [hostname]...
	$progname -U hostname [hostname]...
	$progname -s hostname

	-h: Prints this usage statement and exits.
	-V: Prints version number and exits.
	-u: Person who made the change. Defaults to current user.
	-D: Date when change was made (as YYYY-MM-DD). Defaults to current date.
	-f: File containing change description. Defaults to standard input.
	-U: Undoes most recent change log entry.
	-s: Shows the change log for the specified host." >&2
	exit 2
}

setup () {
	progname=`basename $0`
	VERSION="$progname 2.2"

	while getopts hVu:D:f:Us c; do
		case $c in
			h)	usage;;
			V)	echo $VERSION
				exit 2;;
			u)	changer=$OPTARG;;
			D)	chDate=$OPTARG;;
			f)	chDescFile=$OPTARG;;
			U)	UNDO="TRUE";;
			s)	SHOW="TRUE";;
			\?)	usage;;
		esac
	done
	shift `expr $OPTIND - 1`

	# Need at least one argument.
	[ $# -ge 1 ] || usage

	# No more than one of these should be set.
	[ "X$UNDO" != "X" -a "X$SHOW" != "X" ] && usage

	if [ "X$UNDO" = "X" -a "X$SHOW" = "X" ]; then
		[ "X$changer" = "X" ] && changer=`id -un`
		os=`uname -s`
		if [ "X$os" = "XDarwin" ]; then
			# Look up user in Open Directory.
			odEnt=`dscl localhost -read /Search/Users/$changer RealName`
			[ "X$odEnt" != "Xread: Invalid Path" ] &&
				fname=`echo $odEnt | cut -d" " -f2-`
		else
			# Look up user in passwd file.
			pwEnt=`grep "^${changer}:" /etc/passwd`
			[ "X$pwEnt" != "X" ] &&
				fname=`echo $pwEnt | cut -d":" -f5`
		fi
		[ "X$fname" != "X" ] && changer="$changer ($fname)"

		[ "X$chDate" = "X" ] && chDate=`date +%Y-%m-%d`
		isValidDate $chDate || \
			error "Date format of $chDate is invalid."

		if [ "X$chDescFile" != "X" ]; then
			isValidPathname $chDescFile || \
				error "Filename $chDescFile is invalid."
			[ -f $chDescFile ] || error "Can't read $chDescFile."
		fi
	fi

	hostList=""
	for host in $*; do
		if isValidHostname $host; then
			hostList="$hostList $host"
		else
			warn "Hostname $host is invalid."
		fi
	done

	# Guess at directory to use. E.g., if called as /usr/local/bin/chrec,
	# then $chDir is /usr/local/var/changes/.
	baseDir=`dirname $0`
	cd $baseDir; cd ..; baseDir=`pwd`
	chDir="$baseDir/var/changes"
	# Alternative: hard-code directory.
	#chDir="/usr/local/var/changes"

	bakDir="$chDir/BACKUPS"
	if [ ! -d $bakDir ]; then
		echo "Creating $chDir..."
		mkdir -m 0775 -p $chDir
		mkdir -m 0775 $bakDir
	fi
}

isValidDate () {
	date=$1

	# If the date matches the desired format, we snip out the whole thing
	# and are left with an empty string.
	date=`echo $date | sed 's/^2[0-9][0-9][0-9]-[0-1][0-9]-[0-3][0-9]$//'`

	return `[ "X$date" = "X" ]`
}

isValidPathname () {
	path=$1

	# Pathname should include only slashes, alphanumerics, and a few other
	# kinds of characters. This may be too restrictive, but better safe
	# than sorry.
	#
	# Remove all but valid characters.
	newPath=`echo $path | sed "s/\[^\/a-zA-Z0-9._-+\]//g"`

	# Compare pathname with invalid characters removed to original. If it's
	# the same, we're OK, otherwise return false.
	return `[ "X$path" = "X$newPath" ]`
}

isValidHostname () {
	hostname=$1

	# Hostname should include only alphanumerics, dots, hyphens, and
	# underscores.
	#
	# Remove all but valid characters.
	newHostname=`echo $hostname | sed "s/\[^a-zA-Z0-9._-\]//g"`

	# Compare hostname with invalid characters removed to original. If it's
	# the same, we're OK, otherwise return false.
	return `[ "X$hostname" = "X$newHostname" ]`
}

warn () {
	msg=$*
	echo "$progname: $msg" >&2
}

error () {
	msg=$*
	echo "$progname: $msg" >&2
	exit 2
}

cleanup () {
	status=$1

	# If a temporary change description file was created, remove it.
	[ "X$TEMP" = "X" ] || rm $chDescFile

	exit $status
}

setup $*
echo

if [ "X$SHOW" != "X" ]; then
	for host in $hostList; do
		cat $chDir/$host
	done
elif [ "X$UNDO" != "X" ]; then
	echo "Undoing most recent update to the following change logs:"
	echo
	for host in $hostList; do
		chFile="$chDir/$host"
		bakFile="$bakDir/$host"

		if [ -f $bakFile ]; then
			echo $chFile
			mv $bakDir/$host $chFile
		else
			echo "$chFile: no backup file"
		fi
	done
else
	if [ "X$chDescFile" = "X" ]; then
		# Create temporary change description file from user input.
		TEMP="TRUE"
		chDescFile="$chDir/TEMP.$$"

		echo "Enter description of change, ending with ^D on a line by itself:"
		cat - > $chDescFile
	fi

	echo "Updating the following change logs:"
	echo
	for host in $hostList; do
		chFile="$chDir/$host"
		echo $chFile

		if [ -f $chFile ]; then
			cp -p $chFile $bakDir/$host
		else
			touch $chFile
			chmod 0664 $chFile
		fi

		cat <<EOF >> $chFile
############################################################

$chDate		$changer
----------
`cat $chDescFile`

EOF
	done
fi

echo
cleanup 0
